#!/bin/python
# The SIMPLEX performer network, sheafified!
# (Based on what was presented at the program  kickoff, April 22, 2015)
#
# Copyright (c) 2015, Michael Robinson
# Distribution of unaltered copies permitted for noncommercial use only
# All other uses require express permission of the author
# This software comes with no warrantees express or implied 

import networkx as nx
import pysheaf as ps
import numpy as np
import matplotlib.pyplot as plt

class SIMPLEXSheaf(ps.Sheaf,ps.DirectedGraph):
    def __init__(self,graph,techareas):
        """Construct a DARPA SIMPLEX technical area collaboration sheaf.  
        Vertices represent teams and directed edges represent the desire for one team to collaborate with another.  
        If two teams of different technical areas wish to collaborate, that's considered consistent 
        The stalk over each team is either
          dimension 1 if the team is in one technical area (techareas[i] = 1 or 2)
          dimension 2 if the team is in both technical areas (techareas[i] = 3)
        The stalk over each edge is always two dimensional
        Restriction along an outgoing edge preserves order, while restriction along an incoming edge flips it"""""
        
        sheafcells=[]
        i=0
        for c in graph.cells:
            cofaces=[]
            for cf in c.cofaces: 
                # Construct corestrictions
                if cf.orientation == 1: # Outgoing
                    if techareas[i] == 1:
                        corest=np.matrix([[1],[0]])
                    if techareas[i] == 2:
                        corest=np.matrix([[0],[1]])
                    if techareas[i] == 3:
                        corest=np.matrix([[1,0],[0,1]])                                         
                else: # Incoming
                    if techareas[i] == 1:
                        corest=np.matrix([[0],[1]])
                    if techareas[i] == 2:
                        corest=np.matrix([[1],[0]])
                    if techareas[i] == 3:
                        corest=np.matrix([[0,1],[1,0]])
                cofaces.append(ps.SheafCoface(cf.index,cf.orientation,corest))

            if c.cofaces: # Note edges won't have cofaces
                i+=1
                sheafcells.append(ps.SheafCell(dimension=c.dimension,
                                               compactClosure=c.compactClosure,
                                               cofaces=cofaces))
            else:
                sheafcells.append(ps.SheafCell(dimension=c.dimension,
                                               compactClosure=c.compactClosure,
                                               cofaces=cofaces,
                                               stalkDim=2))

        ps.Sheaf.__init__(self,sheafcells)

# Performers:
team_list=['American University',   # 0
           'Georgetown University', # 1
           'American Museum of Natural History', # 2
           'Columbia University', # 3
           'MIT', # 4
           'QuesTek', # 5
           'Harvard University', # 6
           'John Hopkins University', # 7
           'Stanford University (MINER)', # 8
           'Stanford University (Resident)', #9
           'UCLA', # 10
           'Vanderbilt', # 11
           'Sotera Defense'] # 12

# Distinct collaborations (as of the kickoff meeting on April 22, 2015)
#       AU -> UCLA,Vanderbilt,Baylor,Georgetown
#	Georgetown -> Sotera, AMNH, Stanford Resident, MIT (Demaine), AU, Harvard, Vanderbilt
#	AMNH/UNC -> Harvard, Stanford DeepDive, Georgetown, Stanford MINER
#	Columbia -> Stanford DeepDive, Johns Hopkins TA1, Stanford MINER T1
#	MIT (3D Print) -> Stanford MINER TA1, Vanderbilt, QuesTek, AU
#	QuesTek -> Vanderbilt, Stanford MINER TA1, Stanford DeepDive, MIT (3D Print)
#	Harvard -> UCLA, AMNH/UNC
#	JHU -> Columbia, Stanford DeepDive
#	Stanford MINER -> Columbia, QuesTek, MIT, Stanford Resident, AU, Sotera
#	Stanford Resident -> Georgetown, Stanford MINER
#	UCLA -> AU, Vanderbilt, Harvard
#       Vanderbilt -> AU, Baylor, Georgetown, and Sotera
#	Sotera -> Baylor, Columbia
collab_digraph=[[0,10],[0,11],[0,8],[0,1],
                [1,12],[1,2],[1,9],[1,4],[1,0],[1,6],[1,11],
                [2,6],[2,9],[2,1],[2,8],
                [3,9],[3,7],[3,8],
                [4,8],[4,11],[4,5],[4,0],
                [5,11],[5,8],[5,9],[5,4],
                [6,10],[6,2],
                [7,3],[7,9],
                [8,3],[8,5],[8,4],[8,9],[8,0],[8,12],
                [9,1],[9,8],
                [10,0],[10,11],[10,6],
                [11,0],[11,8],[11,1],[11,12],
                [12,8],[12,3]]

# Draw the directed graph thus far...
plt.figure(1)
G=nx.DiGraph()
G.add_edges_from(collab_digraph)
pos=nx.spring_layout(G)
nx.draw_networkx_nodes(G,pos)
nx.draw_networkx_edges(G,pos)
nx.draw_networkx_labels(G,pos,labels={i:team for i,team in enumerate(team_list)})
plt.savefig('s3a3_digraph.png')

# Technical Areas for each team: 1 = TA1, 2 = TA2, 3 = Both 
techareas=[1, # 0
           1, # 1
           2, # 2
           2, # 3
           2, # 4
           2, # 5
           1, # 6
           3, # 7
           3, # 8
           3, #9
           3, # 10
           1, # 11
           3] # 12

ssh1=SIMPLEXSheaf(ps.DirectedGraph(collab_digraph),techareas)

plt.figure(2)
nx.draw(ssh1.attachDiagram())
plt.savefig('s3a3_attachments.png')

print 'SIMPLEX Program 0-Cohomology has dimension ' + str(ssh1.cohomology(0).shape[1])
if ssh1.cohomology(0).shape[1] == 0:
    print 'This means that some teams from the same technical area want to collaborate'
else:
    print 'All teams want to collaborate ONLY with teams of complementary technical areas'
    
print 'SIMPLEX Program 1-Cohomology has dimension ' + str(ssh1.cohomology(1).shape[1])
if ssh1.cohomology(1).shape[1] > 0:
    print 'This means that there are ' + str(ssh1.cohomology(1).shape[1]) + ' ways groups of teams could mutually collaborate on a joint project'

plt.show()
